/*----------------------------------------------------------------------
  File    : cluster1.c
  Contents: cluster and cluster set management (basic functions)
  Author  : Christian Borgelt
  History : 2001.09.05 file created
            2001.09.11 functions cls_range(), cls_expand()
            2001.09.15 first version of fuzzy c-means completed
            2002.08.07 parsing function completed
            2003.01.30 data normalization code added
            2003.02.01 extended functions (for tuples) added
            2003.02.02 extended parse function added
            2003.02.07 reinitialization added to function cls_range()
            2003.03.09 some initialization moved to separate routines
            2003.03.13 membership degree normalization made optional
            2003.04.10 matrix initialization bug in params() fixed
            2003.05.15 membership normalization modes added
            2003.05.16 noise clustering added
            2003.08.06 bug in mapping binary attributes fixed
            2003.08.11 adapted to new module attmap
            2003.08.12 adapted to new module nstats
            2003.08.14 bug in params() fixed (covariance matrix reading)
            2003.08.15 adapted to new module radfn
            2004.02.16 bug in params() for array of variances fixed
            2004.02.23 bug in function params() fixed
            2004.02.25 adapted to changed radfn, bug in cls_exec() fixed
            2004.02.27 function cls_deletex() added
            2004.03.01 update functions moved to a separate file
            2004.03.17 missing evaluation of cluster weight added
            2004.03.18 normalization parameters added
            2004.04.25 bug in function cls_unscale() fixed
            2004.06.21 reading covariances made more robust
            2004.07.28 bug in function params() (cls_parse()) fixed
            2004.08.12 adapted to new module parse
            2007.01.10 use of att. maps changed, bug in norm() fixed
            2007.11.26 noise cluster configured with separate function
            2007.11.27 membership degree transformation extended
            2008.02.29 redesign for feature weighting completed
            2008.04.02 reading and writing of joint (co)variances added
            2008.04.03 numerical problems solved, consistent matrices
            2008.04.08 auxiliary functions extended (cls_vars() etc.)
            2008.04.11 rescaling removed from covariance matrices
            2008.04.14 cls_unscale() adapted to removed rescaling
            2008.04.25 check for positive eigenvalues added to covmat()
            2010.04.08 special handling of squares removed
            2013.08.16 adapted to definition of type DIMID in matrix.h
            2013.08.30 missing init. of (cov|inv|smp)[0][0] added
            2018.08.18 extended functions (attset/table) removed
            2021.08.23 bug in function "parse()" fixed (clset->nst)
            2022.04.13 bug in function "centers()" fixed (input < 0)
            2022.04.17 bug in function "centers()" fixed (z = (n +...))
----------------------------------------------------------------------*/
#include <stdlib.h>
#include <limits.h>
#include <float.h>
#include <string.h>
#include <assert.h>
#include "cluster.h"
#include "arrays.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE     16          /* block size for cluster array */
#define MINVAR      1e-48       /* minimal variance */
#define MAXVAR      1e+48       /* maximal variance */
#define MINDET      1e-48       /* minimal determinant */
#define MAXDET      1e+48       /* maximal determinant */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static char *nrmnames[] =       /* names of normalization modes */
  { "none", "sum1", "max1", "hard" };

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static CLSET* create (DIMID inpcnt, DIMID clscnt)
{                               /* --- create the base structure */
  CLSET *clset = (CLSET*)malloc(sizeof(CLSET));
  if (!clset) return NULL;      /* create the base structure */
  clset->type     = CLS_CENTER; /* prototypes consist of centers only */
  clset->tnew     = CLS_CENTER;
  clset->inpcnt   = inpcnt;
  clset->clscnt   = clscnt;
  clset->init     = 0;
  clset->cls      = NULL;
  clset->radfn    = rf_cauchy;  /* default radial function: */
  clset->drvfn    = rfd_cauchy; /* \mu(x) = 1 / d^2(\vec{x}, \vec{c}) */
  clset->rfnps[0] = 2;          /* (standard fuzzy clustering) */
  clset->rfnps[1] = 0;
  clset->unit     = 1;
  clset->fwexp    = 0;          /* no feature weighting/selection */
  clset->noise    = 0;          /* do not use a noise cluster */
  clset->ncmsd[0] = 0;
  clset->ncmsd[1] = 0;
  clset->norm     = CLS_SUM1;   /* default normalization: */
  clset->nrmps[0] = 1;          /* u_ij = u_ij^* / sum_k u_{kj}^* */
  clset->nrmps[1] = 0;
  clset->msexp    = 2;          /* use standard fuzzy clustering */
  clset->method   = CLS_ALTOPT; /* with alternating optimization */
  clset->moment   = 0;
  clset->growth   = 1.2;
  clset->shrink   = 0.5;
  clset->maxchg   = 1;
  clset->maxfac   = clset->growth /(clset->growth+1);
  clset->rates[0] =  0.2;
  clset->rates[1] =  0.2;
  clset->rates[2] =  0.2;
  clset->decay[0] = -0.2;
  clset->decay[1] = -0.2;
  clset->decay[2] = -0.2;
  clset->regps[0] = 0;          /* regularization parameters */
  clset->regps[1] = 1;          /* (b, a, s, h, w) */
  clset->regps[2] = 1;
  clset->regps[3] = 0;
  clset->regps[4] = 0;
  clset->steps    = 0;
  clset->resized  = 0;
  clset->nst      = NULL;       /* initialize pointers, */
  clset->vec      = NULL;       /* so that deletion with */
  clset->buf      = NULL;       /* cls_delete() works properly */
  clset->mat      = NULL;
  memset(&clset->eval, 0, sizeof(CLSEVAL));
  return clset;                 /* return the created cluster set */
}  /* create() */

/*--------------------------------------------------------------------*/

static int initcls (CLUSTER *c, DIMID dim)
{                               /* --- init. a cluster prototype */
  assert(c && (dim > 0));       /* check the function arguments */
  c->cov = mat_create(dim, dim);
  c->inv = mat_create(dim, dim);
  c->smp = mat_create(dim, dim);
  c->chv = mat_create(dim, dim);
  c->grv = mat_create(dim, dim);
  if (!c->cov || !c->inv || !c->smp || !c->chv || !c->grv)
    return -1;                  /* create cluster-specific matrices */
  mat_set(c->cov, 0, 0, 1);     /* init. isotropic variance */
  mat_set(c->inv, 0, 0, 1);
  mat_set(c->smp, 0, 0, 1);
  c->ctr = mat_vector(c->cov);  /* get their associated vectors */
  c->dif = mat_vector(c->inv);  /* for easier access */
  c->sum = mat_vector(c->smp);
  c->chc = mat_vector(c->chv);
  c->grc = mat_vector(c->grv);
  return c->dec = 0;            /* return 'ok' */
}  /* initcls() */

/*--------------------------------------------------------------------*/

void cls_combine (CLSET *clset)
{                               /* --- combine (co)variances */
  int     m;                    /* matrix location */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */

  assert(clset                  /* check the function argument */
  &&    (clset->type & CLS_JOINT));
  if      (clset->type & CLS_COVARS) m = MAT_UPPER;
  else if (clset->type & CLS_VARS)   m = MAT_DIAG;
  else if (clset->type & CLS_SIZE)   m = MAT_CORNER;
  else return;                  /* get the matrix location */
  c = clset->cls;               /* get the cluster array */
  mat_mulsx(c[0].cov, c[0].cov, mat_getwgt(c[0].cov), m);
  for (i = 1; i < clset->clscnt; i++) /* sum weighted (co)variances */
    mat_addx(c[0].cov, c[0].cov, mat_getwgt(c[i].cov), c[i].cov, m);
  /* It is assumed that the sum of the cluster weights is 1 and */
  /* thus that no renormalization of the result is necessary.   */
}  /* cls_combine() */

/*--------------------------------------------------------------------*/

void cls_distrib (CLSET *clset)
{                               /* --- distribute (co)variances */
  int     m;                    /* matrix location */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */

  assert(clset                  /* check the function argument */
  &&    (clset->type & CLS_JOINT));
  if      (clset->type & CLS_COVARS)          m = MAT_FULL;
  else if (clset->type & (CLS_VARS|CLS_SIZE)) m = MAT_DIAG;
  else return;                  /* get the matrix location */
  c = clset->cls;               /* and the cluster array */
  for (i = 1; i < clset->clscnt; i++) {
    c[i].var   = c[0].var;      /* traverse the clusters and */
    c[i].size  = c[0].size;     /* copy the isotropic variance, */
    c[i].scale = c[0].scale;    /* the cluster size and the scaling */
    mat_copyx(c[i].cov, c[0].cov, m);
    mat_copyx(c[i].inv, c[0].inv, m);
  }                             /* copy (co)variances and inverses */
}  /* cls_distrib() */

/*--------------------------------------------------------------------*/

static void covmat (CLUSTER *c, int eigen)
{                               /* --- decompose covariance matrix */
  int    r;                     /* result of function calls */
  DIMID  i, n;                  /* loop variable, number of dims. */
  double t, det;                /* buffers for shift, determinant */

  assert(c);                    /* check the function arguments */
  n = mat_rowcnt(c->cov);       /* get the number of dimensions */
  for (t = 0.5*MINVAR, i = 96; --i >= 0; ) {
    if (c->dec) r = 0;          /* do several decomposition trials */
    else {                      /* if decomposition does not exist */
      if (!eigen)               /* do Cholesky decomposition */
        r = mat_chdecom(c->inv, c->cov);
      else {                    /* do eigenvalue decomposition */
        mat_3dred(c->buf, c->buf +n, c->inv, c->cov, MAT_UPPER);
        r = mat_3dqli(c->dif, c->inv, c->buf, c->buf +n, n, 256);
        if (r == 0) {           /* if successfully decomposed */
          mat_evsort(c->dif, c->inv, 1);
          r = (c->dif[0] >= MINVAR) ? 0 : -1;
        }                       /* check for positive eigenvalues */
      }                         /* (positive definite matrix) */
    }
    if (r == 0) {               /* if decomposition was successful */
      if (eigen) {              /* if eigenvector decomposition */
        det    = vec_prod(c->dif, n);
        c->var = ((det >= MINDET) && (det <= MAXDET))
               ? pow(det,               1.0/(double)n)
               : exp(vec_sumlog(c->dif, n) /(double)n); }
      else {                    /* if Cholesky decomposition */
        det    = mat_chdet(c->inv);
        c->var = ((det >= MINDET) && (det <= MAXDET))
               ? pow(det,            1.0/(double)n)
               : exp(mat_chlogd(c->inv) /(double)n);
      }                         /* compute the isotropic variance */
      if (c->var > MINVAR) break;
    }                           /* check for a proper value */
    mat_diaadds(c->cov, t += t);/* if the decomposition failed, */
  }                             /* shift the eigenvalues */
  if (i < 0) {                  /* if decomposition failed totally, */
    c->var = 1;                 /* set a unit covariance matrix */
    mat_init(c->cov, MAT_UNIT, NULL);
    mat_init(c->inv, MAT_UNIT, NULL);
    if (eigen) { while (--n >= 0) c->dif[i] = 1; }
  }                             /* set unit eigenvalues if necessary */
  c->dec = -1;                  /* set the decomposition flag */
}  /* covmat() */

/*--------------------------------------------------------------------*/

static void diamat (CLUSTER *c)
{                               /* --- diagonal covariance matrix */
  DIMID  i, n;                  /* loop variable, number of dims. */
  double t, det;                /* buffers for shift, determinant */

  assert(c);                    /* check the function argument */
  n = mat_rowcnt(c->cov);       /* get the number of dimensions */
  for (t = MINVAR, i = 97; --i >= 0; ) {
    det = mat_diaprod(c->cov);  /* compute the determinant */
    c->var = ((det >= MINDET) && (det <= MAXDET))
           ? pow(det,            1.0/(double)n)
           : exp(mat_dialog(c->cov) /(double)n);
    if (c->var > MINVAR) break; /* compute the isotropic variance */
    mat_diaadds(c->cov, t);     /* if the computation failed, */
    if (i < 96) t += t;         /* shift the eigenvalues and */
  }                             /* double the increment value */
  if (i >= 0) return;           /* if sucessful computation, abort */
  c->var = 1;                   /* otherwise set a unit cluster size */
  mat_diasetx(c->cov, 1);       /* and a unit covariance matrix */
}  /* diamat() */

/*----------------------------------------------------------------------
The above functions do a safe decomposition of the covariance matrix or
a safe computation of the isotropic variance of a cluster by shifting
the eigenvalues in case of a failure. This is repeated with larger
and larger shifts and only if after a shift by a fairly large value
(MINVAR *2^96) the decomposition/variance computation still fails, a
unit matrix is set for the covariance matrix (and its decomposition).
----------------------------------------------------------------------*/

void cls_vars (CLSET *clset, int eigen)
{                               /* --- compute isotropic variances */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */

  assert(clset);                /* check the function argument */
  clset->eigen = 0;             /* clear the eigenvalue decomp. flag */
  i = (clset->type & CLS_JOINT) ? 1 : clset->clscnt;
  c = clset->cls;               /* get the cluster array */
  if (clset->type & CLS_COVARS){/* if full covariance matrix */
    if ((clset->fwexp >  0)     /* if feature weighting */
    &&  (clset->fwexp != 1))    /* (exponent or alternative), */
      eigen = -1;               /* eigenvalue decomposition is needed */
    clset->eigen = eigen;       /* note the eigenvalue decomp. flag */
    while (--i >= 0)            /* traverse the clusters and */
      covmat(c+i, eigen); }     /* decompose the covariance matrices */
  else if (clset->type & CLS_VARS) {   /* if only variances, */
    while (--i >= 0)            /* traverse the clusters and */
      diamat(c+i); }            /* handle diagonal matrices */
  else {                        /* if only isotropic variances */
    while (--i >= 0) {          /* traverse the clusters */
      c[i].var = mat_get(c[i].cov, 0, 0);
      if (!(c[i].var >= MINVAR))/* get isotropic variance */
        mat_set(c[i].cov, 0, 0, c[i].var = MINVAR);
      if (!(c[i].var <= MAXVAR))
        mat_set(c[i].cov, 0, 0, c[i].var = MAXVAR);
    }                           /* clamp the isotropic variance */
  }                             /* to a reasonable range */
}  /* cls_vars() */

/*--------------------------------------------------------------------*/

void cls_gauge (CLSET *clset)
{                               /* --- measure the cluster sizes */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  t;                    /* buffer for computations */

  assert(clset);                /* check the function argument */
  i = (clset->type & CLS_JOINT) ? 1 : clset->clscnt;
  c = clset->cls;               /* get the cluster array */
  if ((clset->fwexp <= 0)       /* if no feature weighting, */
  ||  (clset->fwexp == 1)       /* the isotropic variance is the size */
  || !(clset->type & (CLS_COVARS|CLS_VARS))) {
    while (--i >= 0)            /* traverse the clusters and */
      c[i].scale = c[i].var; }  /* copy the isotropic variance */
  else {                        /* if feature weighting */
    assert(clset->eigen         /* check for eigenvalue decomposition */
    ||   !(clset->type & CLS_COVARS));
    while (--i >= 0) {          /* traverse the clusters */
      t = (clset->type & CLS_COVARS)
        ? vec_sumrec(c[i].dif, clset->inpcnt) : mat_diarec(c[i].cov);
      c[i].scale = (t > MINVAR) ? 1/t : MAXVAR;
    }                           /* compute the cluster size */
  }                             /* the sum of the feature weights */
}  /* cls_gauge() */

/*--------------------------------------------------------------------*/

void cls_resize (CLSET *clset, int decomp)
{                               /* --- (re)scale the clusters */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */

  assert(clset);                /* check the function argument */
  i = (clset->type & CLS_JOINT) ? 1 : clset->clscnt;
  c = clset->cls;               /* get the cluster array */
  if (clset->type & CLS_SIZE) { /* if adaptable cluster size, */
    clset->resized = 1;         /* set flag for resized clusters */
    while (--i >= 0) {          /* traverse the clusters */
      c[i].size  = c[i].scale;  /* simple copy the cluster size */
      c[i].scale = 1;           /* and set a unit scaling factor */
    } }
  else {                        /* if the cluster size is fixed */
    while (--i >= 0) {          /* traverse the clusters */
      c[i].scale = c[i].size /c[i].scale; /* compute scaling factor */
      if ((c[i].scale == 1) || !(clset->type & CLS_COVARS) || !decomp)
        continue;               /* check whether scaling is needed */
      if (clset->eigen)         /* rescale the eigenvalues */
        vec_muls (c[i].dif, clset->inpcnt, c[i].dif, c[i].scale);
      else mat_mulsx(c[i].inv, c[i].inv, sqrt(c[i].scale), MAT_LOWER);
    }                           /* rescale the decomposition */
  }                             /* of the covariance matrix */
  if (!(clset->type & (CLS_COVARS|CLS_VARS)) && clset->resized) {
    i = (clset->type & CLS_JOINT) ? 1 : clset->clscnt;
    while (--i >= 0)            /* traverse the clusters */
      mat_diasetx(c[i].cov, mat_get(c[i].cov, 0, 0));
  }                             /* (re)set the diagonal */
}  /* cls_resize() */

/*--------------------------------------------------------------------*/

void cls_ftwgts (CLSET *clset)
{                               /* --- compute feature weights */
  DIMID   i, k, n;              /* loop variables, number of dims. */
  CLUSTER *c;                   /* to traverse the clusters */
  double  x, v, sum;            /* buffers for computations */
  double  off;                  /* transformation offset */

  assert(clset);                /* check the function argument */
  i = (clset->type & CLS_JOINT) ? 1 : clset->clscnt;
  c = clset->cls;               /* get the cluster array */
  n = clset->inpcnt;            /* and the number of dimensions */
  x = clset->fwexp;             /* get the feature weight exponent */

  /* --- compute feature weights --- */
  if ((x > 0) && (x != 1)       /* if to do feature weighting */
  &&  (clset->type & (CLS_COVARS|CLS_VARS))) {
    assert(clset->eigen         /* check for eigenvalue decomposition */
    ||   !(clset->type & CLS_COVARS));
    while (--i >= 0) {          /* traverse the clusters */
      sum = 1/c[i].scale;       /* compute the scaling factor */
      for (k = n; --k >= 0; ) { /* traverse the dimensions */
        v = (clset->type & CLS_COVARS)
          ? c[i].dif[k] : mat_get(c[i].cov, k, k);
        c[i].dif[k] = sum / ((v > 0) ? v : MINVAR);
      }                         /* compute the feature weights */
      if (x > 1) continue;      /* if normal exponent, skip extension */
      vec_copy (clset->buf, c[i].dif, n); /* sort inverse eigenvalues */
      dbl_qsort(clset->buf, (size_t)n, -1);
      for (sum = v = clset->buf[k = 0]; ++k < n; ) {
        v = clset->buf[k];      /* traverse the sorted eigenvalues */
        if (v *(1 +x*(double)k) <= x *(sum +v)) break;
        sum += v;               /* find the number of features */
      }                         /* that receive a positive weight */
      /* Note that if two (or more) degrees of membership are equal, */
      /* then either all of them or none of them satisfy the above   */
      /* condition for a non-zero transformed degree of membership.  */
      if (!(sum > 0)) sum = 1;  /* make sure the sum does not vanish */
      off = x *(v = 1/(double)(1-x));   /* compute trans. parameters */
      sum = (1 +x *(double)(k-1)) *v /sum;
      for (k = n; --k >= 0; ) { /* traverse the inverse eigenvalues */
        v = sum *c[i].dif[k] -off;
        c[i].dif[k] = (v > 0) ? v : 0;
      }                         /* compute the feature weights, */
    } }                         /* possibly setting some to zero */

  /* --- compute inverse eigenvalues --- */
  else if (clset->eigen) {      /* if eigenvalue decomposition */
    while (--i >= 0) {          /* traverse the clusters */
      sum = 1/c[i].scale;       /* compute the scaling factor */
      for (k = n; --k >= 0; ) {
        v = c[i].dif[k];        /* traverse the eigenvalues */
        c[i].dif[k] = sum / ((v > 0) ? v : MINVAR);
      }                         /* compute inverse eigenvalues */
    } }                         /* in a safe way */
  else if (clset->type & CLS_COVARS)
    { }                         /* if Cholesky decomp., do nothing */
  else if (clset->type & CLS_VARS) {
    while (--i >= 0) {          /* traverse the clusters */
      sum = 1/c[i].scale;       /* compute the scaling factor */
      for (k = n; --k >= 0; ) { /* traverse the variances */
        v = mat_get(c[i].cov, k, k);
        c[i].dif[k] = sum / ((v > 0) ? v : MINVAR);
      }                         /* compute inverse variances */
    } }                         /* in a safe way */
  else if ((clset->type & CLS_SIZE) || clset->resized) {
    while (--i >= 0) {          /* traverse the clusters */
      v = mat_get(c[i].cov, 0, 0);
      c[i].dif[0] = (1/c[i].scale) /((v > 0) ? v : MINVAR);
    }                           /* invert the isotropic variance */
  }                             /* in a safe way */
}  /* cls_ftwgts() */

/*--------------------------------------------------------------------*/

void cls_invert (CLSET *clset)
{                               /* --- compute inverse matrices */
  DIMID   i, k, n;              /* loop variables, buffers */
  CLUSTER *c;                   /* to traverse the clusters */
  double  t, x, b;              /* buffers for computations */

  assert(clset);                /* check the function argument */
  i = (clset->type & CLS_JOINT) ? 1 : clset->clscnt;
  c = clset->cls;               /* get the cluster array and */
  x = clset->fwexp;             /* the feature weight exponent */

  /* --- if feature weights --- */
  if ((x > 0) && (x != 1)       /* if to do feature weighting */
  &&  (clset->type & (CLS_COVARS|CLS_VARS))) {
    assert(clset->eigen         /* check for eigenvalue decomposition */
    ||   !(clset->type & CLS_COVARS));
    b = (x < 1) ? (1-x)/(1+x) : 0; /* compute transform. parameter */
    n = clset->inpcnt;          /* get the number of dimensions */
    while (--i >= 0) {          /* traverse the clusters */
      k = n;                    /* if normal feature weight exponent, */
      if (x > 1) {              /* simply adapt inverse eigenvalues */
        while (--k >= 0) c[i].dif[k] = pow(c[i].dif[k], x); }
      else {                    /* if alternative to an exponent */
        while (--k >= 0) {      /* traverse the dimensions */
          t = (b *(c[i].dif[k] -1) +1) *c[i].dif[k];
          c[i].dif[k] = (t > 0) ? t : 0;
        }                       /* transform the feature weights */
      }                         /* with alternative formula */
      if (!(clset->type & CLS_COVARS)) mat_diaset(c[i].inv, c[i].dif);
      else mat_mulmdmt(c[i].inv, c[i].inv, c[i].dif);
    } }                         /* compute the distance norm matrix */

  /* --- if normal (co)variances --- */
  else {                        /* if no feature weighting */
    if (clset->eigen)           /* if eigenvalue decomposition */
      while (--i >= 0) {        /* multiply eigenvalues and -vectors */
        mat_mulmdmt(c[i].inv, c[i].inv, c[i].dif); }
    else if (clset->type & CLS_COVARS) {
      while (--i >= 0) {        /* if Cholesky decomposition */
        mat_tr2sym(c[i].cov, c[i].cov, MAT_UPPER);
        mat_chinv (c[i].inv, c[i].inv);
        mat_tr2sym(c[i].inv, c[i].inv, MAT_UPPER);
      } }                       /* compute inverse covariance matrix */
    else if (clset->type & CLS_VARS)
      while (--i >= 0)          /* if only variances, set inverses */
        mat_diaset (c[i].inv, c[i].dif);
    else if (clset->resized) {  /* if clusters were resized, */
      while (--i >= 0)          /* set the inverse variance */
        mat_diasetx(c[i].inv, c[i].dif[0]);
    }                           /* (which is also stored only in */
  }                             /* c[i].dif first for consistency) */
  clset->resized = 0;           /* clear flag for resized clusters */
}  /* cls_invert() */

/*--------------------------------------------------------------------*/

void cls_reinit (CLSET *clset)
{                               /* --- reinit. aggregation matrices */
  int     m;                    /* matrix flags */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */

  if      (clset->type & CLS_COVARS)          m = MAT_UPPER;
  else if (clset->type & (CLS_VARS|CLS_SIZE)) m = MAT_DIAG;
  else                                        m = 0;
  m |= MAT_VECTOR|MAT_WEIGHT;   /* get the initialization mode */
  c  = clset->cls;              /* and the cluster array */
  for (i = 0; i < clset->clscnt; i++) {
    c[i].dec = 0;               /* traverse the clusters */
    mat_init(c[i].smp, m, NULL);/* clear the decomposition flag */
  }                             /* and reinitialize the matrices */
}  /* cls_reinit() */

/*----------------------------------------------------------------------
  Cluster Set Functions
----------------------------------------------------------------------*/

CLSET* cls_create (DIMID inpcnt, DIMID clscnt)
{                               /* --- create a set of clusters */
  DIMID   i;                    /* loop variable */
  CLSET   *clset;               /* created set of clusters */
  CLUSTER *c;                   /* to traverse the clusters */

  assert((inpcnt > 0)           /* check the function arguments */
  &&     (clscnt > 0));         /* (at least one dim., one cluster) */
  clset = create(inpcnt,clscnt);/* create the base structure */
  if (!clset) return NULL;      /* and initialize the variables */
  clset->cls = c = (CLUSTER*)malloc((size_t)clscnt *sizeof(CLUSTER));
  if (!clset->cls) { cls_delete(clset); return NULL; }
  for (i = 0; i < clscnt; i++)  /* clear pointers for cleanup */
    c[i].cov = c[i].inv = c[i].smp = c[i].chv = c[i].grv = NULL;
  clset->vec = (double*)malloc(3 *(size_t)inpcnt *sizeof(double));
  if (!clset->vec) { cls_delete(clset); return NULL; }
  clset->buf = clset->vec +inpcnt; /* create the numeric vectors */
  clset->nst = nst_create(inpcnt); /* and the input statistics */
  clset->mat = mat_create(inpcnt, inpcnt);
  if (!clset->nst || !clset->mat)  { cls_delete(clset); return NULL; }
  for (i = 0; i < clscnt; i++) {
    if (initcls(c+i, inpcnt) != 0) { cls_delete(clset); return NULL; }
    c[i].buf = clset->buf;      /* initialize the clusters and */
    c[i].mat = clset->mat;      /* note the computation buffers */
    memset(c[i].val, 0, sizeof(c[i].val));
  }                             /* clear evaluations */
  clset->setup = 0;             /* cluster set is not yet set up */
  return clset;                 /* return created set of clusters */
}  /* cls_create() */

/*--------------------------------------------------------------------*/

void cls_delete (CLSET *clset)
{                               /* --- delete a set of clusters */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */

  assert(clset);                /* check the function argument */
  if (clset->mat) mat_delete(clset->mat);
  if (clset->nst) nst_delete(clset->nst);
  if (clset->vec) free(clset->vec);
  if (clset->cls) {             /* if there is a cluster array */
    c = clset->cls;             /* traverse the clusters */
    for (i = 0; i < clset->clscnt; i++) {
      if (c[i].cov) mat_delete(c[i].cov);
      if (c[i].inv) mat_delete(c[i].inv);
      if (c[i].smp) mat_delete(c[i].smp);
      if (c[i].chv) mat_delete(c[i].chv);
      if (c[i].grv) mat_delete(c[i].grv);
    }                           /* delete all matrices */
    free(clset->cls);           /* and the cluster array */
  }
  free(clset);                  /* delete the base structure */
}  /* cls_delete() */

/*--------------------------------------------------------------------*/

CLSET* cls_clone (CLSET *clset)
{                               /* --- clone a cluster set */
  DIMID   i;                    /* loop variable */
  CLUSTER *c, *d;               /* to traverse the clusters */
  CLSET   *clone;               /* created clone of the cluster set */

  assert(clset);                /* check the function arguments */
  clone = (CLSET*)malloc(sizeof(CLSET));
  if (!clone) return NULL;      /* create a new cluster structure */
  memcpy(clone, clset, sizeof(CLSET));
  clone->cls = (CLUSTER*)malloc((size_t)clone->clscnt *sizeof(CLUSTER));
  clone->vec = (double*)malloc(3*(size_t)clone->inpcnt *sizeof(double));
  clone->nst = nst_clone(clset->nst);  /* create the numeric vectors */
  clone->mat = mat_clone(clset->mat);  /* and other objects */
  if (!clone->cls || !clone->vec || !clone->nst || !clone->mat) {
    cls_delete(clone); return NULL; }
  clone->buf = clone->vec +clone->inpcnt;
  memcpy(clone->vec, clset->vec, (size_t)clone->inpcnt *sizeof(double));
  d = clone->cls;                /* clear pointers for cleanup */
  for (i = 0; i < clone->clscnt; i++)
    d[i].cov = d[i].inv = d[i].smp = d[i].chv = d[i].grv = NULL;
  c = clset->cls;                /* traverse the clusters */
  for (i = 0; i < clone->clscnt; i++) {
    d[i].cov   = mat_clonex(c[i].cov, MAT_FULL|MAT_VECTOR|MAT_WEIGHT);
    d[i].inv   = mat_clonex(c[i].inv, MAT_FULL|MAT_VECTOR|MAT_WEIGHT);
    d[i].smp   = mat_clonex(c[i].smp, MAT_FULL|MAT_VECTOR|MAT_WEIGHT);
    d[i].chv   = mat_clonex(c[i].chv, MAT_FULL|MAT_VECTOR|MAT_WEIGHT);
    d[i].grv   = mat_clonex(c[i].grv, MAT_FULL|MAT_VECTOR|MAT_WEIGHT);
    if (!d[i].cov || !d[i].inv || !d[i].smp || !d[i].chv || !d[i].grv) {
      cls_delete(clone); return NULL; }
    d[i].ctr   = mat_vector(d[i].cov);
    d[i].dif   = mat_vector(d[i].inv);
    d[i].sum   = mat_vector(d[i].smp);
    d[i].chc   = mat_vector(d[i].chv);
    d[i].grc   = mat_vector(d[i].grv);
    d[i].buf   = clone->buf;    /* get associated vectors */
    d[i].mat   = clone->mat;    /* for easier access */
    d[i].ok    = c[i].ok;
    d[i].dec   = c[i].dec;
    d[i].var   = c[i].var;
    d[i].size  = c[i].size;
    d[i].scale = c[i].scale;
    d[i].d2    = c[i].d2;
    d[i].msd   = c[i].msd;
    memcpy(d[i].val, c[i].val, 4 *sizeof(double));
  }
  return clone;                 /* return the created clone */
}  /* cls_clone() */

/*--------------------------------------------------------------------*/

void cls_value (CLSET *clset, DIMID index, double value)
{                               /* --- set an input value */
  clset->vec[index] = (value -nst_offset(clset->nst, index))
                             *nst_factor(clset->nst, index);
}  /* cls_value() */

/*--------------------------------------------------------------------*/

void cls_type (CLSET *clset, int type, double size)
{                               /* --- set the cluster type */
  DIMID i;                      /* loop variable */

  assert(clset);                /* check the function arguments */
  clset->setup = 0;             /* cluster set needs to be set up */
  if (type >= 0) {              /* if a new cluster type is given */
    if (clset->inpcnt <= 1) type &= ~(CLS_VARS|CLS_COVARS);
    if (type & CLS_COVARS)  type |=   CLS_VARS;
    if ((clset->clscnt < 2) || !(type & (CLS_COVARS|CLS_VARS|CLS_SIZE)))
      type &= ~CLS_JOINT;       /* make cluster type consistent */
    clset->tnew = type & CLS_ALL;
  }                             /* note the new cluster type */
  if (size < 0) return;         /* check the initial cluster size */
  for (i = 0; i < clset->clscnt; i++)
    clset->cls[i].size = size;  /* store the given cluster size */
  clset->resized = -1;          /* set flag for resized clusters */
}  /* cls_type() */

/*--------------------------------------------------------------------*/

void cls_radfn (CLSET *clset, RADFN *radfn, DERIVFN *drvfn,
                const double *params)
{                               /* --- set the membership function */
  assert(clset);                /* check the function arguments */
  clset->setup = 0;             /* cluster set needs to be set up */
  if (radfn) clset->radfn = radfn;  /* note the radial function */
  if (drvfn) clset->drvfn = drvfn;  /* and its derivative (if given) */
  if (!params) return;          /* check for function parameters */
  clset->rfnps[0] = params[0];  /* note the function parameters and */
  clset->rfnps[1] = params[1];  /* compute the unit integral factor */
}  /* cls_radfn() */

/*--------------------------------------------------------------------*/

void cls_norm (CLSET *clset, int mode, const double *params)
{                               /* --- set normalization mode */
  assert(clset);                /* check the function arguments */
  clset->setup = 0;             /* cluster set needs to be set up */
  if (mode >= 0)                /* if the norm. mode is valid, */
    clset->norm = mode;         /* note the normalization mode */
  if (!params) return;          /* if normalization parameters */
  clset->nrmps[0] = params[0];  /* are given, copy them */
  clset->nrmps[1] = params[1];  /* to the cluster set */
}  /* cls_norm() */

/*--------------------------------------------------------------------*/

void cls_msexp (CLSET *clset, double msexp)
{                               /* --- set the membership exponent */
  assert(clset);                /* check the function arguments */
  clset->setup = 0;             /* cluster set needs to be set up */
  if (msexp == 0) {             /* if to use a corresponding value */
    if      (clset->nrmps[0] >= 1) msexp = 1 /clset->nrmps[0] +1;
    else if (clset->nrmps[0] >  0) msexp =    clset->nrmps[0];
    else                           msexp = 2;
  }                             /* compute the membership exponent */
  clset->msexp = msexp;         /* store the membership exponent */
}  /* cls_msexp() */

/*--------------------------------------------------------------------*/

int cls_setup (CLSET *clset)
{                               /* --- set up a cluster set */
  int     k;                    /* cluster flags */
  DIMID   i, n;                 /* loop variables */
  CLUSTER *c;                   /* to traverse the clusters */
  double  t;                    /* buffer for computations */

  assert(clset);                /* check the function argument */
  if (clset->setup) return 0;   /* check whether already set up */
  t = clset->noise;             /* compute raw noise cluster ms.deg. */
  clset->ncmsd[0] = (t >= 0) ? t : clset->radfn(t*t, clset->rfnps);
  clset->unit     = clset->radfn((double)-clset->inpcnt, clset->rfnps);
  if (clset->fwexp < 0) clset->fwexp = 0;

  /* --- normalize centers --- */
  c = clset->cls;               /* get the cluster array */
  if (clset->method & CLS_UNIT) {
    for (i = 0; i < clset->clscnt; i++)
      vec_unitlen(c[i].ctr, c[i].ctr, clset->inpcnt);
  }                             /* make center of unit length */

  /* --- compute cluster weights --- */
  if (!(clset->type & CLS_WEIGHT)) {
    t = 1/(double)clset->clscnt;  /* if no adaptable cluster weight */
    for (i = 0; i < clset->clscnt; i++)
      mat_setwgt(c[i].cov, t); }/* set same weight for all clusters */
  else {                        /* if the clusters have a weight */
    t = 0;                      /* init. the cluster weight sum */
    for (i = 0; i < clset->clscnt; i++)
      t += mat_getwgt(c[i].cov);/* sum the cluster weights */
    t = (t > 0) ? 1/t : 1;      /* compute the normalization factor */
    for (i = 0; i < clset->clscnt; i++)
      mat_mulwgt(c[i].cov, t);  /* normalize the cluster weights */
  }                             /* (ensure a unit sum) */

  /* --- gauge cluster sizes --- */
  if ((clset->tnew & CLS_JOINT) && !(clset->type & CLS_JOINT)) {
    clset->type |= CLS_JOINT;   /* if (co)variances are to be joined, */
    cls_combine(clset);         /* combine the (co)variances */
  }                             /* then decompose cov. matrices and */
  cls_vars (clset, 0);          /* compute the isotropic variances */
  cls_gauge(clset);             /* and measure the cluster sizes */

  /* --- adapt type and rescale clusters --- */
  k = clset->type & clset->tnew & (CLS_COVARS|CLS_VARS);
  n = (clset->type & CLS_JOINT) ? 1 : clset->clscnt;
  if (k & CLS_COVARS) {         /* if covars to covars */
    for (i = 0; i < n; i++) {   /* traverse the clusters */
      c[i].msd = c[i].move = 0; /* initialize the membership degree */
      if ((clset->tnew & CLS_SIZE) || (c[i].size <= 0)) {
        c[i].size = c[i].scale; c[i].scale = 1; }
      else {                    /* if rescaling is needed */
        c[i].scale = t = c[i].size /c[i].scale;
        if (t == 1) continue;   /* compute the scaling factor */
        if (clset->eigen) vec_muls(c[i].dif, clset->inpcnt, c[i].dif,t);
        else mat_mulsx(c[i].inv, c[i].inv, sqrt(t), MAT_LOWER);
      }                         /* rescale the decomposition */
    } }                         /* of the covariance matrix */
  else {                        /* if no covariances */
    for (i = 0; i < n; i++) {   /* traverse the clusters */
      c[i].msd = c[i].move = 0; /* initialize the membership degree */
      if ((clset->tnew & CLS_SIZE) || (c[i].size <= 0)) {
        c[i].size  = c[i].scale; c[i].scale = 1; }
      else                      /* if rescaling is needed, */
        c[i].scale = c[i].size /c[i].scale; /* compute scaling factor */
      if (!(k & CLS_VARS))      /* distribute isotropic variance */
        mat_diasetx(c[i].cov, c[i].size);
      if (!(clset->type & CLS_COVARS)) continue;
      mat_crop(c[i].cov, MAT_DIAG);
      mat_crop(c[i].inv, MAT_DIAG);
    }                           /* clear covariances if necessary */
  }                             /* (keep cov and inv consistent) */

  /* --- finalize setup --- */
  clset->type = clset->tnew;    /* set the new cluster type */
  if ((clset->tnew & CLS_COVARS)/* if covariances have been added */
  && !(k           & CLS_COVARS)) {
    for (i = 0; i < n; i++) {   /* traverse the clusters */
      mat_crop(c[i].cov, MAT_DIAG);
      mat_crop(c[i].inv, MAT_DIAG);
    }                           /* clear the covariances and compute */
    clset->type &= ~CLS_COVARS; /* feature weights/inverse matrices */
  }                             /* as for only variances */
  cls_ftwgts(clset);            /* compute the feature weights */
  cls_invert(clset);            /* and the inverse matrices */
  clset->type = clset->tnew;    /* set the new cluster type again */
  if (clset->type & CLS_JOINT)  /* if joint (co)variances, */
    cls_distrib(clset);         /* distribute joint (co)variances */
  cls_reinit(clset);            /* reinit. aggregation matrices */
  clset->setup = 1;             /* cluster set has been set up */
  return 0;                     /* return 'ok' */
}  /* cls_setup() */

/*--------------------------------------------------------------------*/

double cls_sqrdist (CLSET *clset, DIMID clsid, const double *vec)
{                               /* --- compute raw squared distance */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to access the cluster */

  if (vec) nst_norm(clset->nst, vec, clset->vec);
  vec = clset->vec;             /* scale given vector to the buffer */
  c = clset->cls +clsid;        /* get the cluster referred to */
  for (i = 0; i < clset->inpcnt; i++)      /* compute difference */
    c->dif[i] = vec[i] -c->ctr[i];         /* vector to center and */
  return vec_sqrlen(c->dif,clset->inpcnt); /* return squared distance */
}  /* cls_sqrdist() */

/*--------------------------------------------------------------------*/

DIMID cls_exec (CLSET *clset, const double *vec, double *msdegs)
{                               /* --- compute degrees of membership */
  DIMID   i, k, best;           /* loop variables, cluster index */
  CLUSTER *c;                   /* to traverse the clusters */
  double  *z, *b;               /* to access the center and buffer */
  double  d, x;                 /* buffers for computations */
  double  sum, max;             /* sum/maximum of membership degrees */

  assert(clset && clset->setup);/* check the function arguments */
  if (vec) nst_norm(clset->nst, vec, clset->vec);
  vec = clset->vec;             /* scale given vector to the buffer */

  /* --- compute raw degrees of membership --- */
  for (k = 0; k < clset->clscnt; k++) {
    c = clset->cls +k;          /* traverse the clusters */
    z = c->ctr; b = c->dif;     /* get center and difference vector */
    for (i = 0; i < clset->inpcnt; i++)
      b[i] = vec[i] -z[i];      /* compute diff. vector to center */
    if      (clset->type & CLS_COVARS) /* -- if to use covariances */
      d = mat_mulvmv(c->inv,b); /* compute Mahalanobis distance */
    else if (clset->type & CLS_VARS)   /* -- if to use variances */
      d = mat_mulvdv(c->inv,b); /* compute coord. weighted distance */
    else                        /* -- if to use isotropic variance */
      d = vec_sqrlen(b, clset->inpcnt) *mat_get(c->inv, 0, 0);
    if      (  d <  0 ) d = 0;  /* check the computed distance */
    else if (!(d >= 0)) d = MAXVAR;

    /* Here d contains the squared and scaled distance between */
    /* the data point and the center of the current cluster c. */
    c->d2  = d;                 /* note distance (for eval. measures) */
    c->msd = clset->radfn(d, clset->rfnps);
    if ((clset->type & CLS_NORM) && (clset->unit > 0)) {
      d = sqrt(pow(c->var, (double)clset->inpcnt));
      if (d > 0) c->msd *= clset->unit /d;
      else       c->msd  = 0;   /* normalize membership degree */
    }                           /* for a probability distribution */
    if (clset->type & CLS_WEIGHT)   /* if to use adaptable weights, */
      c->msd *= mat_getwgt(c->cov); /* incorporate the weight/prior */
    if (!(c->msd >= 0))         /* ensure a valid membership degree */
      c->msd = 0;               /* (note that !(x >= 0) is true */
  }                             /* even if x is not a valid number) */

  /* --- transform degrees of membership --- */
  if (clset->norm != CLS_HARD){ /* if not hard/crisp clustering */
    d = clset->nrmps[1];        /* get the transformation parameters */
    x = clset->nrmps[0];        /* (offset and exponent) */
    if (x == 0) {               /* if to use corresponding value */
      if      (clset->msexp >= 1) x = 1 /(clset->msexp -1);
      else if (clset->msexp >  0) x =     clset->msexp;
      else                        x = 1;
    }                           /* compute the corresponding value */
    if (x <  1) {               /* -- if alternative fuzzifier */
      for (k = 0; k < clset->clscnt; k++)   /* get and sort the */
        clset->buf[k] = clset->cls[k].msd;  /* membership degrees */
      dbl_qsort(clset->buf, (size_t)k, -1); /* descendingly */
      for (sum = d = clset->buf[i = 0]; ++i < k; ) {
        d = clset->buf[i];      /* traverse the membership degrees */
        if (d *(1 +x*(double)i) <= x *(sum +d)) break;
        sum += d;               /* find number of clusters with */
      }                         /* a positive degree of membership */
      /* Note that if two (or more) feature weights are equal, then */
      /* either all or none of them satisfy the above condition for */
      /* a non-zero transformed feature weight. */
      if (sum <= 0) sum = 1;    /* make sure sum does not vanish */
      d = x *(max = 1/(double)(1-x));  /* compute trans. parameters */
      x = (1 +x *(double)(i-1)) *max /sum;
      for (k = 0; k < clset->clscnt; k++) {
        clset->cls[k].msd = x *clset->cls[k].msd -d;
      } }                       /* transform the membership degrees */
    else {                      /* -- if heuristic transformation */
      if (x != 1) {             /* if not unit exponent */
        for (k = 0; k < clset->clscnt; k++)
          clset->cls[k].msd = pow(clset->cls[k].msd, x);
      }                         /* raise ms. degrees to given power */
      if (d > 0) {              /* if to use an offset */
        max = -INFINITY;        /* find maximal degree of membership */
        for (k = 0; k < clset->clscnt; k++)
          if (clset->cls[k].msd > max) max = clset->cls[k].msd;
        d *= max;               /* compute (relative) offset */
        for (k = 0; k < clset->clscnt; k++)
          clset->cls[k].msd -= d;
      }                         /* subtract the computed offset */
    }                           /* from all membership degrees */
  }  /* if (clset->norm != CLS_HARD) .. */

  /* --- normalize degrees of membership --- */
  sum = 0; best = 0; max = -INFINITY;
  for (k = 0; k < clset->clscnt; k++) {
    c = clset->cls +k;          /* traverse the clusters */
    if (!(c->msd >= 0)) c->msd = 0;
    if (c->msd >= max) { max = c->msd; best = k; }
    sum += c->msd;              /* sum degrees of membership */
  }                             /* and determine their maximum */
  if      (clset->norm == CLS_HARD) {  /* if to do hard clustering */
    for (k = 0; k < clset->clscnt; k++)
      clset->cls[k].msd = 0;    /* clear all membership degrees */
    if (max < clset->ncmsd[0])  /* if the maximum is not high enough, */
      clset->ncmsd[1] = 1;      /* assign pattern to noise cluster */
    else {                      /* if the maximum is high enough */
      clset->ncmsd[1] = 0; clset->cls[best].msd = 1;
    } }                         /* assign pattern to best cluster */
  else if (clset->norm == CLS_MAX1) {  /* if to normalize to max. 1 */
    if (clset->ncmsd[0] > max)  /* compare to membership degree */
      max = clset->ncmsd[0];    /* to noise cluster */
    max = (max > 0) ? 1/max :1; /* compute the normalization factor */
    for (k = 0; k < clset->clscnt; k++)
      clset->cls[k].msd *= max; /* compute rel. membership degrees */
    clset->ncmsd[1] = clset->ncmsd[0] *max; }
  else if (clset->norm == CLS_SUM1) {  /* if to normalize to sum 1 */
    sum += clset->ncmsd[0];     /* add m.s. degree to noise cluster */
    sum = (sum > 0) ? 1/sum :1; /* compute the normalization factor */
    for (k = 0; k < clset->clscnt; k++)
      clset->cls[k].msd *= sum; /* compute rel. membership degrees */
    clset->ncmsd[1] = clset->ncmsd[0] *sum;
  }                             /* set m.s. degree to noise cluster */

  /* --- set the result --- */
  if (msdegs) {                 /* if a result array is given */
    for (k = 0; k < clset->clscnt; k++)
      msdegs[k] = clset->cls[k].msd;
  }                             /* copy the normalized membership */
  return best;                  /* return index of best cluster */
}  /* cls_exec() */             /* (highest degree of membership) */

/*--------------------------------------------------------------------*/

void cls_clsdel (CLSET *clset, DIMID index)
{                               /* --- delete a cluster */
  int     i;                    /* loop variable */
  CLUSTER *c;                   /* cluster to delete */
  double  w;                    /* sum of cluster weights */

  assert(clset                  /* check the function arguments */
  &&    (index >= 0) && (index < clset->clscnt));
  c = clset->cls +index;        /* get the cluster to delete */
  w = 1.0 -mat_getwgt(c->cov);
  if (c->cov) mat_delete(c->cov);
  if (c->inv) mat_delete(c->inv);
  if (c->smp) mat_delete(c->smp);
  if (c->chv) mat_delete(c->chv);
  if (c->grv) mat_delete(c->grv);  /* delete matrices */
  while (++index < clset->clscnt)
    clset->cls[index-1] = clset->cls[index];
  clset->clscnt -= 1;           /* reduce the number of clusters */
  if (w == 1) return;           /* check whether norm. is needed */
  w = (w > 0) ? 1/w : 1;        /* compute normalization factor */
  for (i = 0; i < clset->clscnt; i++)  /* normalize the */
    mat_mulwgt(clset->cls[i].cov, w);  /* cluster weights */
}  /* cls_clsdel() */

/*--------------------------------------------------------------------*/

void cls_clsmrg (CLSET *clset, DIMID idx1, DIMID idx2)
{                               /* --- merge two clusters */
  int     i;                    /* loop variable */
  CLUSTER *c1, *c2;             /* clusters to merge */
  double  w1, w2, sum;          /* cluster weights and their sum */

  assert(clset                  /* check the function arguments */
  &&    (idx1 >= 0) && (idx2 < clset->clscnt)
  &&    (idx2 >= 0) && (idx2 < clset->clscnt));
  c1  = clset->cls +idx1; w1 = mat_getwgt(c1->cov);
  c2  = clset->cls +idx2; w2 = mat_getwgt(c2->cov);
  sum = w1 +w2;                 /* get the cluster to merge */
  w1 /= sum; w2 /= sum;         /* compute their relative weights, */
  mat_setwgt(c1->cov, sum);     /* shift weight to first cluster, */
  mat_setwgt(c2->cov, 0);       /* and merge their centers */
  for (i = 0; i < clset->inpcnt; i++)
    c1->ctr[i] = w1*c1->ctr[i] +w2*c2->ctr[i];
  mat_init(c1->cov, MAT_UNIT, NULL);
  mat_init(c1->inv, MAT_UNIT, NULL);
  c1->dec = 1;                  /* reinit. cov. matrix and inverse */
  cls_clsdel(clset, idx2);      /* delete the merged cluster */
}  /* cls_clsmrg() */

/*--------------------------------------------------------------------*/

void cls_sclcov (CLSET *clset)
{                               /* --- scale (co)variances */
  int     m;                    /* matrix flags */
  DIMID   i;                    /* loop variables */
  CLUSTER *c;                   /* to traverse the clusters */

  assert(clset);                /* check the function argument */
  if (!(clset->type & (CLS_COVARS|CLS_VARS|CLS_SIZE)))
    return;                     /* check whether scaling is needed */
  m = (clset->type & CLS_COVARS) ? MAT_UPPER : MAT_DIAG;
  c = clset->cls;               /* get cluster array and matrix pos. */
  for (i = 0; i < clset->clscnt; i++) {
    mat_mulsx(c[i].cov, c[i].cov, c[i].scale, m);
    c[i].var  *= c[i].scale;    /* rescale the covariance matrices, */
    c[i].size *= c[i].scale;    /* the isotropic variance, */
    c[i].scale = 1;             /* and the cluster size and */
  }                             /* finally set a unit scale */
}  /* cls_sclcov() */

/*--------------------------------------------------------------------*/

void cls_unscale (CLSET *clset)
{                               /* --- remove attribute scaling */
  DIMID   i, k, n, m;           /* loop variables */
  CLUSTER *c;                   /* to traverse the clusters */
  double  t;                    /* temporary buffer */

  assert(clset);                /* check the function argument */
  if (!(clset->type & CLS_SIZE)) {
    if (!clset->setup) {        /* if cluster set is not set up */
      cls_vars  (clset, 0);     /* compute the isotropic variances, */
      cls_gauge (clset);        /* measure the cluster sizes, */
      cls_resize(clset, 0);     /* and resize the clusters */
    }
    cls_sclcov(clset);          /* scale the covariance matrices */
  }
  for (i = 0; i < clset->clscnt; i++) /* unscale center coordinates */
    nst_denorm(clset->nst, clset->cls[i].ctr, clset->cls[i].ctr);
  t = nst_factor(clset->nst, 0);
  for (k = clset->inpcnt; --k > 0; )
    if (nst_factor(clset->nst, k) != t)
      break;                    /* check for different scales */
  if (k >= 0) {                 /* if there are different scales */
    if (clset->type & (CLS_COVARS|CLS_VARS)) {
      clset->type |= CLS_SIZE;  /* size are no longer uniform */
      clset->tnew |= CLS_SIZE;  /* (since the dimensions */
    }                           /* are scaled differently) */
    if (!(clset->type & (CLS_COVARS|CLS_VARS))) {
      clset->type |= CLS_VARS;  /* set the flag for variances */
      clset->tnew |= CLS_VARS;  /* (both old and new type) */
      i = (clset->type & CLS_JOINT) ? 1 : clset->clscnt;
      for (c = clset->cls +i; --i >= 0; ) {
        --c; mat_diasetx(c->cov, mat_get(c->cov, 0, 0)); }
    }                           /* set attribute specific variances */
  }                             /* (as they will differ afterwards) */
  if (clset->type & (CLS_COVARS|CLS_VARS|CLS_SIZE)) {
    m = (clset->type & (CLS_COVARS|CLS_VARS)) ? clset->inpcnt : 1;
    n = (clset->type &  CLS_JOINT) ? 1 : clset->clscnt;
    for (c = clset->cls +n; --n >= 0; ) {
      (--c)->size = -1;         /* traverse the clusters */
      for (i = m; --i >= 0; ) { /* traverse the matrix rows */
        k = (clset->type & CLS_COVARS) ? clset->inpcnt : i+1;
        while (--k >= i) {      /* traverse the matrix columns */
          t = nst_factor(clset->nst, i) * nst_factor(clset->nst, k);
          mat_set(c->cov, i, k, mat_get(c->cov, i, k) /t);
        }                       /* compute the scaling factor */
      }                         /* and unscale the elements */
    }                           /* of the covariance matrix */
  }                             /* finally set scaling to identity */
  nst_scale(clset->nst, -1, 0, 1);
  clset->setup = 0;             /* cluster set needs to be set up */
}  /* cls_unscale() */

/*--------------------------------------------------------------------*/

static void descov (CLSET *clset, CLUSTER *c, char *indent, FILE *file)
{                               /* --- describe shape and size */
  DIMID i, k, n;                /* loop variables, number of dims. */

  n = clset->inpcnt;            /* get the number of dimensions */
  if (clset->type & CLS_COVARS){/* if full covariance matrix */
    for (k = 0; k < n; k++) {   /* traverse the matrix rows */
      fprintf(file, ",\n%s            ", indent);
      fprintf(file, "  [% .16g", mat_get(c->cov, 0, k));
      for (i = 0; ++i <= k; )   /* start a new line for each row */
        fprintf(file, ", % .16g", mat_get(c->cov, i, k));
      fputc(']', file);         /* print the covariances */
    }                           /* row by row (lower triangle) */
    if (clset->fwexp > 0) {     /* if to print feature weights */
      fprintf(file, "\n%s              /*", indent);
      for (k = 0; k < n; k++) { /* traverse the matrix rows */
        if (c->dif[k] <= 0) break;
        if (k > 0) fprintf(file, "\n%s                ", indent);
        fprintf(file, " %.16g:", c->dif[k]);
        for (i = 0; i < n; i++) /* start a new line for each row */
          fprintf(file, " % .16g", mat_get(c->inv, i, k));
      }                         /* print the eigenvalues (weights) */
      fputs(" */", file);       /* and the corresponding eigenvectors */
    } }                         /* finally terminate the comment */
  else if (clset->type & CLS_VARS) { /* if only variances */
    fprintf(file, ",\n%s            ", indent);
    fprintf(file, "  [%.16g", mat_get(c->cov, 0, 0));
    for (i = 0; ++i < n; )      /* traverse the dimensions */
      fprintf(file, ", %.16g", mat_get(c->cov, i, i));
    fputc(']', file);           /* print a list of variances */
    if (clset->fwexp > 0) {     /* if to print feature weights */
      fprintf(file, "\n%s              /*", indent);
      for (i = 0; i < n; i++)   /* traverse the dimensions again */
        fprintf(file, " %.16g", c->dif[i]);
      fputs(" */", file);       /* print the feature weights and */
    } }                         /* finally terminate the comment */
  else if (clset->type & CLS_SIZE) { /* if only cluster size */
    fprintf(file, ",\n%s            ", indent);
    fprintf(file, "  [%.16g]", mat_get(c->cov, 0, 0));
  }                             /* print the isotropic variance */
}  /* descov() */

/*--------------------------------------------------------------------*/

int cls_desc (CLSET *clset, FILE *file, int mode, int maxlen)
{                               /* --- describe a set of clusters */
  DIMID   i, k;                 /* loop variables */
  int     len, l;              /* loop variables for comments */
  CLUSTER *c;                   /* to traverse the clusters */
  char    *indent = "";         /* indentation string */

  assert(clset && file);        /* check the function arguments */
  len = (maxlen > 0) ? maxlen -2 : 70;

  /* --- print a header (as a comment) --- */
  if (mode & CLS_TITLE) {       /* if the title flag is set */
    fputs("/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fputs("\n  cluster set\n", file);
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (asa comment) */

  if (!(mode & CLS_RBFNET))     /* if not part of an RBF network */
    fputs("clusters = {\n", file); /* start cluster set description */
  indent = "  ";                /* indent the description itself */

  /* --- print scaling parameters --- */
  nst_desc(clset->nst, file, indent);

  /* --- print the function description --- */
  fputs(indent,        file);   /* print the indentation */
  fputs("function = ", file);   /* and the function name */
  fputs((clset->radfn == rf_gauss) ? "gauss" : "cauchy", file);
  fprintf(file, "(%.16g,%.16g)", clset->rfnps[0], clset->rfnps[1]);
  if (clset->type & CLS_NORM)   /* print the function parameters */
    fputs(", normalized",file); /* and normalization flag if needed */
  fputs(";\n", file);           /* terminate the output line */

  /* --- print the noise cluster parameter --- */
  if (clset->noise != 0) {      /* if noise cluster parameter given */
    fputs(indent, file);        /* write the indentation */
    fprintf(file, "noise    = %.16g;\n", clset->noise);
  }                             /* print noise cluster parameter */

  /* --- print the normalization mode --- */
  fputs(indent,        file);   /* write the indentation and */
  fputs("normmode = ", file);   /* the normalization mode indicator */
  fputs(nrmnames[clset->norm], file);
  if ((clset->nrmps[0] != 1)    /* if not standard parameters, */
  ||  (clset->nrmps[1] != 0))   /* print normalization parameters */
    fprintf(file, "(%.16g,%.16g)", clset->nrmps[0], clset->nrmps[1]);
  fputs(";\n", file);           /* finally terminate the line */

  /* --- print the feature weight exponent --- */
  if (clset->fwexp > 0) {       /* if a feature weight exponent given */
    fputs(indent, file);        /* write the indentation */
    fprintf(file, "fwexp    = %.16g;\n", clset->fwexp);
  }                             /* write the feature weight exponent */

  /* --- print the membership exponent --- */
  if (clset->msexp != 2) {      /* if it does not have default value */
    fputs(indent, file);        /* write the indentation */
    fprintf(file, "msexp    = %.16g;\n", clset->msexp);
  }                             /* write the feature weight exponent */

  /* --- print cluster parameters --- */
  fputs(indent,         file);  /* write the indentation and */
  fputs("params   = {", file);  /* start the parameter section */
  if (clset->fwexp > 0) {       /* if to print feature weights, */
    cls_vars  (clset, 1);       /* compute the isotropic variances, */
    cls_gauge (clset);          /* measure the cluster sizes, */
    cls_resize(clset, 1);       /* resize the clusters, and */
    cls_ftwgts(clset);          /* compute the feature weights */
  }                             /* (extended parameter output) */
  c = clset->cls;               /* traverse the clusters */
  for (i = 0; i < clset->clscnt; c++, i++) {
    if (i > 0)                  /* start a new output line */
      fprintf(file, ",\n%s            ", indent);
    fprintf(file, "{ [% .16g", c->ctr[0]);
    for (k = 0; ++k < clset->inpcnt; )
      fprintf(file, ", % .16g", c->ctr[k]);
    fputc(']', file);           /* print the cluster center and */
    if (!(clset->type & CLS_JOINT) /* shape and size parameters */
    &&   (clset->type & (CLS_COVARS|CLS_VARS|CLS_SIZE)))
      descov(clset, c, indent, file);
    if (clset->type & CLS_WEIGHT) {
      fputc(',', file);         /* print separating comma */
      if ((clset->type & (CLS_VARS|CLS_COVARS))
      || !(clset->type & (CLS_VARS|CLS_COVARS|CLS_SIZE)))
        fprintf(file, "\n%s             ", indent);
      fprintf(file, " %.16g", mat_getwgt(c->cov));
    }                           /* print the cluster weight */
    fputs(" }", file);          /* terminate the cluster description */
  }

  /* --- print joint parameters --- */
  if ((clset->type & CLS_JOINT) /* if joint shape and size parameters */
  &&  (clset->type & (CLS_COVARS|CLS_VARS|CLS_SIZE)))
    descov(clset, clset->cls, indent, file);
  if (!(clset->type & CLS_SIZE) /* if uniform cluster size */
  && (clset->cls[0].size != 1)) /* that differs from the default 1 */
    fprintf(file, ",\n%s              %.16g ",
                  indent, clset->cls[0].size);
  fputs("};\n", file);         /* terminate the parameter section */
  if (clset->fwexp > 0)         /* if to print feature weights, */
    cls_invert(clset);          /* compute inverse matrices */

  if (!(mode & CLS_RBFNET))     /* if not part of an RBF network */
    fputs("};\n", file);        /* terminate the description */

  /* --- print additional information (as a comment) --- */
  if (mode & CLS_INFO) {        /* if the add. info. flag is set */
    fputs("\n/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file, "\n  number of inputs  : %"DIMID_FMT, clset->inpcnt);
    fprintf(file, "\n  number of clusters: %"DIMID_FMT, clset->clscnt);
    fputc('\n', file);          /* print additional cluster info. */
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* (as a comment) */
  }
  return ferror(file) ? -1 : 0; /* return the write status */
}  /* cls_desc() */

/*--------------------------------------------------------------------*/

int cls_cdesc (MATRIX *ctrs, NSTATS *nst,
               FILE *file, int mode, int maxlen)
{                               /* --- describe a set of clusters */
  DIMID  i, k;                  /* loop variables */
  DIMID  c;                     /* number of clusters */
  DIMID  m;                     /* number of dimensions/vector elems. */
  double *vec;                  /* to access the center vectors */
  int    len, l;                /* loop variables for comments */
  char   *indent = "";          /* indentation string */

  assert(ctrs && file);         /* check the function arguments */
  len = (maxlen > 0) ? maxlen -2 : 70;

  /* --- print a header (as a comment) --- */
  if (mode & CLS_TITLE) {       /* if the title flag is set */
    fputs("/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fputs("\n  cluster set\n", file);
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (asa comment) */

  fputs("clusters = {\n", file);/* start the cluster set description */
  indent = "  ";                /* and indent the description itself */

  /* --- print scaling parameters --- */
  if (nst) nst_desc(nst, file, indent);

  /* --- print function and normalization --- */
  fputs(indent, file);          /* print the function name */
  fputs("function = cauchy(2,0);\n", file);
  fputs(indent, file);          /* print the normalization mode */
  fputs("normmode = hard;\n", file);

  /* --- print cluster parameters --- */
  fputs(indent,         file);  /* write the indentation and */
  fputs("params   = {", file);  /* start the parameter section */
  c = mat_rowcnt(ctrs);         /* get the number of clusters */
  m = mat_colcnt(ctrs);         /* and the number of dimensions */
  for (i = 0; i < c; i++) {     /* traverse the clusters */
    if (i > 0)                  /* start a new output line */
      fprintf(file, ",\n%s            ", indent);
    vec = mat_row(ctrs, i);     /* get the next cluster center */
    fprintf(file, "{ [% .16g", vec[0]);
    for (k = 0; ++k < m; )
      fprintf(file, ", % .16g", vec[k]);
    fputc(']', file);           /* print the cluster center and */
    fputs(" }", file);          /* terminate the cluster description */
  }
  fputs(" };\n", file);         /* terminate the parameter section */
  fputs("};\n", file);          /* terminate the description */

  /* --- print additional information (as a comment) --- */
  if (mode & CLS_INFO) {        /* if the add. info. flag is set */
    fputs("\n/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file,"\n  number of inputs  : %"DIMID_FMT,mat_colcnt(ctrs));
    fprintf(file,"\n  number of clusters: %"DIMID_FMT,mat_rowcnt(ctrs));
    fputc('\n', file);          /* print additional cluster info. */
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* (as a comment) */
  }
  return ferror(file) ? -1 : 0; /* return the write status */
}  /* cls_cdesc() */

/*----------------------------------------------------------------------
  Cluster Set Parse Functions
----------------------------------------------------------------------*/
#ifdef CLS_PARSE

static int scales (NSTATS **pnst, DIMID inpcnt, SCANNER *scan)
{                               /* --- get the input scalings */
  assert(pnst && scan);         /* check the function arguments */
  if (*pnst) return 0;          /* check whether scalings exist */
  if ((scn_token(scan) != T_ID) /* check whether 'scales' follows */
  ||  (strcmp(scn_value(scan), "scales") != 0))
    return 0;                   /* if not, abort the function */
  *pnst = nst_parse(scan, inpcnt);
  if (!*pnst) return 1;         /* parse the scaling parameters */
  return 0;                     /* get the number of dimensions */
}  /* scales() */

/*--------------------------------------------------------------------*/

static int function (CLSET *clset, SCANNER *scan)
{                               /* --- parse the function description */
  const char *s;                /* buffer for token value */
  double     p;                 /* buffer for a parameter */

  assert(scan);                 /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for "function" */
  ||  (strcmp(scn_value(scan), "function") != 0))
    return 0;                   /* if there is none, abort */
  SCN_NEXT(scan);               /* consume the "function" keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  if (scn_token(scan) != T_ID)  /* check for a keyword/identifier */
    SCN_ERROR(scan, E_STREXP, "cauchy");
  s = scn_value(scan);          /* check for a function name */
  if      (strcmp(s, "cauchy") == 0) {
    if (clset) { clset->radfn = rf_cauchy; clset->drvfn = rfd_cauchy; }}
  else if (strcmp(s, "gauss")  == 0) {
    if (clset) { clset->radfn = rf_gauss;  clset->drvfn = rfd_gauss; } }
  else SCN_ERROR(scan, E_STREXP, "cauchy");
  SCN_NEXT(scan);               /* and consume it */
  SCN_CHAR(scan, '(');          /* consume '(' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  p = strtod(scn_value(scan), NULL);
  if (p <= 0) SCN_ERROR(scan, E_NUMBER);
  if (clset) clset->rfnps[0] = p;  /* get the first parameter, */
  SCN_NEXT(scan);               /* i.e., the distance exponent */
  SCN_CHAR(scan, ',');          /* consume the separating ',' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  p = strtod(scn_value(scan), NULL);
  if (p < 0) SCN_ERROR(scan, E_NUMBER);
  if (clset) clset->rfnps[1] = p;  /* get the second parameter, */
  SCN_NEXT(scan);               /* i.e., the scaling factor */
  SCN_CHAR(scan, ')');          /* consume the closing ')' */
  if (scn_token(scan) == ',') { /* if a comma follows */
    SCN_NEXT(scan);             /* consume ',' */
    if ((scn_token(scan) != T_ID)
    ||  (strcmp(scn_value(scan), "normalized") != 0))
      SCN_ERROR(scan, E_STREXP, "normalized");
    if (clset) clset->type |= CLS_NORM;
    SCN_NEXT(scan);             /* get the normalization flag */
  }                             /* consume the "normalized" keyword */
  SCN_CHAR(scan, ';');          /* consume the closing ';' */
  return 0;                     /* return 'ok' */
}  /* function() */

/*--------------------------------------------------------------------*/

static int noise (CLSET *clset, SCANNER *scan)
{                               /* --- parse the noise cluster param. */
  assert(scan);                 /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for "noise" */
  ||  (strcmp(scn_value(scan), "noise") != 0)) {
    if (clset) { clset->noise = 0; } return 0; }
  SCN_NEXT(scan);               /* consume the "noise" keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  if (clset) clset->noise = strtod(scn_value(scan), NULL);
  SCN_NEXT(scan);               /* get the noise cluster parameter */
  SCN_CHAR(scan, ';');          /* consume the closing ';' */
  return 0;                     /* return 'ok' */
}  /* noise() */

/*--------------------------------------------------------------------*/

static int normmode (CLSET *clset, SCANNER *scan)
{                               /* --- parse the normalization mode */
  int        i;                 /* loop variable for norm. modes */
  const char *s;                /* buffer for token value */
  double     n;                 /* buffer for a parameter */

  assert(scan);                 /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for "normmode" */
  ||  (strcmp(scn_value(scan), "normmode") != 0)) {
    if (clset) { clset->norm = CLS_NONE; } return 0; }
  SCN_NEXT(scan);               /* consume the "normmode" keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  if (scn_token(scan) != T_ID)  /* check for a keyword/identifier */
    SCN_ERROR(scan, E_STREXP, "sum1");
  s = scn_value(scan);          /* check for a norm. mode name */
  for (i = sizeof(nrmnames)/sizeof(*nrmnames); --i >= 0; )
    if (strcmp(s, nrmnames[i]) == 0) {
      if (clset) { clset->norm = i; } break; }
  if (i < 0) SCN_ERROR(scan, E_STREXP, nrmnames[CLS_SUM1]);
  SCN_NEXT(scan);               /* decode the normalization mode */
  if (scn_token(scan) == '(') { /* if a parenthesis follows, */
    SCN_NEXT(scan);             /* consume the '(' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    n = strtod(scn_value(scan), NULL);
    if (n < 0)                    SCN_ERROR(scan, E_NUMBER);
    if (clset) clset->nrmps[0] = n;
    SCN_NEXT(scan);             /* get the 1st normalization param. */
    SCN_CHAR(scan, ',');        /* consume the separating comma */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    n = strtod(scn_value(scan), NULL);
    if (n < 0)                    SCN_ERROR(scan, E_NUMBER);
    if (clset) clset->nrmps[1] = n;
    SCN_NEXT(scan);             /* get the 2nd normalization param. */
    if (clset->nrmps[0] == 0) { /* if to use a corresponding value */
      SCN_CHAR(scan, ',');      /* consume the separating comma */
      if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
      n = strtod(scn_value(scan), NULL);
      if (n <= 0)                   SCN_ERROR(scan, E_NUMBER);
      if (clset) clset->nrmps[1] = n;
      SCN_NEXT(scan);           /* get 1st normalization parameter */
    }                           /* and consume the number */
    SCN_CHAR(scan, ')');        /* consume the closing ')' */
  }
  SCN_CHAR(scan, ';');          /* consume the closing ';' */
  return 0;                     /* return 'ok' */
}  /* normmode() */

/*--------------------------------------------------------------------*/

static int fwexp (CLSET *clset, SCANNER *scan)
{                               /* --- parse the feature weight exp. */
  double x;                     /* buffer for the exponent */

  assert(scan);                 /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for "fwexp" */
  ||  (strcmp(scn_value(scan), "fwexp") != 0)) {
    if (clset) { clset->fwexp = 0; } return 0; }
  SCN_NEXT(scan);               /* consume the "fwexp" keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  x = strtod(scn_value(scan), NULL);
  if (x < 0)                    SCN_ERROR(scan, E_NUMBER);
  if (clset) clset->fwexp = x;  /* get the feature weight exponent */
  SCN_NEXT(scan);               /* and consume the number */
  SCN_CHAR(scan, ';');          /* consume the closing ';' */
  return 0;                     /* return 'ok' */
}  /* fwexp() */

/*--------------------------------------------------------------------*/

static int msexp (CLSET *clset, SCANNER *scan)
{                               /* --- parse the membership exp. */
  double x;                     /* buffer for the exponent */

  assert(scan);                 /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for "fwexp" */
  ||  (strcmp(scn_value(scan), "msexp") != 0)) {
    if (clset) { clset->msexp = 2; } return 0; }
  SCN_NEXT(scan);               /* consume the "fwexp" keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  x = strtod(scn_value(scan), NULL);
  if (x < 0)                    SCN_ERROR(scan, E_NUMBER);
  if (clset) clset->msexp = x;  /* get the feature weight exponent */
  SCN_NEXT(scan);               /* and consume the number */
  SCN_CHAR(scan, ';');          /* consume the closing ';' */
  return 0;                     /* return 'ok' */
}  /* msexp() */

/*--------------------------------------------------------------------*/

static int covars (CLSET *clset, CLUSTER *c, SCANNER *scan, int known)
{                               /* --- read (co)variances */
  DIMID  i, k;                  /* loop variables */
  double t;                     /* buffer for a (co)variance */

  SCN_CHAR(scan, '[');          /* consume '[' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  t = strtod(scn_value(scan), NULL);
  if (t < 0) SCN_ERROR(scan, E_NUMBER);
  SCN_NEXT(scan);               /* get and check a variance */
  if (t < MINVAR) t = MINVAR;   /* clamp the variance */
  if (t > MAXVAR) t = MAXVAR;   /* to a reasonable range */
  mat_set(c->cov, 0, 0, t);     /* set the (first) variance */

  while (!known) {              /* if the cluster type is unknown */
    clset->type |= CLS_SIZE;    /* default: size is adaptable */
    if (clset->inpcnt < 2)      /* if there is only one dimension, */
      break;                    /* only a size is needed */
    if (scn_token(scan) == ','){/* if a comma follows, then variances */
      clset->type |= CLS_VARS; break; }
    SCN_CHAR(scan, ']');        /* consume ']' */
    if (scn_token(scan) != ',') /* if no comma follows, there is */
      break;                    /* only an isotropic variance */
    SCN_NEXT(scan);             /* consume ',' */
    if (scn_token(scan) == '[') /* if a '[' follows, there is */
      clset->type |= CLS_COVARS;       /* a covariance matrix */
    scn_back(scan); break;      /* push back the scanned '[' */
  }                             /* (it is scanned again below) */

  if (clset->type & CLS_COVARS){/* if to read covariances */
    if (known) { SCN_CHAR(scan, ']'); }  /* consume the closing ']' */
    for (i = 0; ++i < clset->inpcnt; ) {
      SCN_CHAR(scan, ',');      /* consume separating ',' */
      SCN_CHAR(scan, '[');      /* consume opening '[' */
      for (k = 0; k <= i; k++){ /* traverse the columns */
        if (k > 0) { SCN_CHAR(scan, ','); }  /* consume ',' */
        if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
        t = strtod(scn_value(scan), NULL);
        if ((k == i) && (t < 0))      SCN_ERROR(scan, E_NUMBER);
        SCN_NEXT(scan);         /* get and consume a (co)variance */
        if (k == i) {           /* if this is a variance */
          if (t < MINVAR) t = MINVAR;
          if (t > MAXVAR) t = MAXVAR;
        }                       /* clamp it to a reasonable range */
        mat_set(c->cov, k, i, t);
      }                         /* set the (co)variance */
      SCN_CHAR(scan, ']');      /* consume closing ']' */
    } }
  else if (clset->type & CLS_VARS) {  /* if there are variances */
    for (i = 0; ++i < clset->inpcnt; ) {
      SCN_CHAR(scan, ',');      /* consume separating ',' */
      if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
      t = strtod(scn_value(scan), NULL);
      if (t < 0) SCN_ERROR(scan, E_NUMBER);
      SCN_NEXT(scan);           /* get and consume a variance */
      if (t < MINVAR) t = MINVAR;  /* clamp the variance */
      if (t > MAXVAR) t = MAXVAR;  /* to a reasonable range */
      mat_set(c->cov, i, i, t); /* store the variance in the diagonal */
    }
    SCN_CHAR(scan, ']'); }      /* consume closing ']' */
  else if (clset->type & CLS_SIZE)
    if (known) { SCN_CHAR(scan, ']'); }  /* consume closing ']' */
  return 0;                     /* return 'ok' */
}  /* covars() */

/*--------------------------------------------------------------------*/

static int params (CLSET *clset, SCANNER *scan)
{                               /* --- parse a set of clusters */
  int     e;                    /* error status, function result */
  DIMID   i, k;                 /* loop variables */
  DIMID   vsz    = 0;           /* size of the cluster vector */
  DIMID   inpcnt = 0;           /* dimension of (input) data space */
  CLUSTER *c;                   /* to access the clusters */
  double  t, *v = NULL;         /* buffers (for reallocation) */

  assert(clset && scan);        /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for cluster parameters */
  ||  (strcmp(scn_value(scan), "params") != 0))
    SCN_ERROR(scan, E_STREXP, "params");
  SCN_NEXT(scan);               /* consume the "params" keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  while (1) {                   /* cluster read loop */
    SCN_CHAR(scan, '{');        /* consume '{' */
    if (clset->clscnt >= vsz) { /* if the cluster vector is full */
      vsz += (vsz > BLKSIZE) ? vsz >> 1 : BLKSIZE;
      c = (CLUSTER*)realloc(clset->cls, (size_t)vsz *sizeof(CLUSTER));
      if (!c) SCN_ERROR(scan, E_NOMEM);
      clset->cls = c;           /* enlarge the cluster vector */
    }

    /* --- read cluster center --- */
    SCN_CHAR(scan, '[');        /* consume opening '[' */
    if (clset->inpcnt <= 0) {   /* if the dimension is yet unknown */
      clset->inpcnt = 0;        /* initialize the dimension */
      while (1) {               /* coordinate read loop */
        if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
        if (clset->inpcnt >= inpcnt) { /* if coord. vector is full */
          inpcnt += (inpcnt > BLKSIZE) ? inpcnt >> 1 : BLKSIZE;
          v = (double*)realloc(clset->vec,
                               (size_t)inpcnt*sizeof(double));
          if (!v) SCN_ERROR(scan, E_NOMEM);
          clset->vec = v;       /* enlarge the coordinate vector */
        }                       /* and set the new vector */
        clset->vec[clset->inpcnt++] = strtod(scn_value(scan), NULL);
        SCN_NEXT(scan);         /* get and consume the coordinate */
        if (scn_token(scan) != ',') break;
        SCN_NEXT(scan);         /* if at end of coord. list, abort */
      }                         /* the loop, otherwise consume ',' */
    }
    c = clset->cls +clset->clscnt;
    if (initcls(c, clset->inpcnt) != 0)
      SCN_ERROR(scan, E_NOMEM); /* initialize the cluster prototype */
    mat_init(c->cov, MAT_UNIT, NULL);  /* and its covariance matrix */
    if (inpcnt > 0) {           /* if the dimension was unknown */
      vec_copy(c->ctr, clset->vec, clset->inpcnt);
      inpcnt = 0; }             /* copy the read cluster center */
    else {                      /* if the dimension was known */
      for (i = 0; i < clset->inpcnt; i++) {
        if (i > 0) { SCN_CHAR(scan, ','); }  /* consume ',' */
        if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
        c->ctr[i] = strtod(scn_value(scan), NULL);
        SCN_NEXT(scan);         /* traverse the dimensions and */
      }                         /* get and consume the coordinates */
    }                           /* of the center vector */
    SCN_CHAR(scan, ']');        /* consume closing ']' */

    /* --- read (co)variances --- */
    e = 0;                      /* default: no error */
    if (clset->clscnt > 0) {    /* if the cluster type is known */
      if (clset->type & (CLS_COVARS|CLS_VARS|CLS_SIZE)) {
        SCN_CHAR(scan, ',');    /* consume separating comma */
        e = covars(clset, c, scan, -1);
      } }                       /* read the (co)variances */
    else if (scn_token(scan) == ',') {
      SCN_NEXT(scan);           /* if no comma, then only centers */
      if (scn_token(scan) == '[')  /* otherwise, if a '[' follows */
        e = covars(clset, c, scan, 0);
      else scn_back(scan);      /* read (co)variances or */
    }                           /* push back the token */
    if (e < 0) return e;        /* check for a parse error */
    c->size = -1;               /* invalidate the cluster size */

    /* --- read cluster weight --- */
    if ((clset->clscnt <= 0) && (scn_token(scan) == ','))
      clset->type |= CLS_WEIGHT;/* check for a cluster weight */
    if (clset->type & CLS_WEIGHT) {
      SCN_CHAR(scan, ',');      /* consume ',' */
      if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
      t = strtod(scn_value(scan), NULL);
      if (t < 0)                    SCN_ERROR(scan, E_NUMBER);
      SCN_NEXT(scan);           /* get and consume the weight */
      mat_setwgt(c->cov, t);    /* and store it in the cluster */
    }

    SCN_CHAR(scan, '}');        /* consume '}' */
    clset->clscnt++;            /* count the cluster */
    k = scn_token(scan);        /* if at end of cluster list, */
    if (k == '}') break;        /* abort the cluster read loop */
    if (k == ',') {             /* if a comma follows, */
      SCN_NEXT(scan);           /* consume ',' */
      if (scn_token(scan) != '{') { scn_back(scan); break; }
    }                           /* if another cluster follows, */
  }                             /* repeat the read loop */

  /* --- read joint (co)variances --- */
  if (scn_token(scan) == ',') { /* if additional parameters */
    SCN_NEXT(scan);             /* consume ',' */
    if (scn_token(scan) != '[') /* if no joint (co)variances, */
      scn_back(scan);           /* push back the token */
    else if (clset->type & (CLS_COVARS|CLS_VARS|CLS_SIZE))
      SCN_ERROR(scan, E_CHREXP, '{');
    else {                      /* if joint (co)variances, */
      clset->type |= CLS_JOINT; /* set joint parameter flag */
      e = covars(clset, clset->cls, scan, 0);
      if (e < 0) return e;      /* read the joint (co)variances */
    }                           /* and check for a parse error */
  }                             /* (default: adaptable size) */

  /* --- read a fixed size --- */
  t = 1;                        /* get default fixed size */
  if (scn_token(scan) == ',') { /* if a comma follows, */
    SCN_NEXT(scan);             /* consume ',' */
    if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
    t = strtod(scn_value(scan), NULL);
    if (t < 0) SCN_ERROR(scan, E_NUMBER);
    SCN_NEXT(scan);             /* read and check a cluster size */
    clset->type &= ~CLS_SIZE;   /* remove the adaptable size flag */
  }                             /* (the size is now fixed) */
  if (!(clset->type & CLS_SIZE)) {
    for (c = clset->cls +(i = clset->clscnt); --i >= 0; )
      (--c)->size = t;          /* if there is a fixed size, */
  }                             /* store the size in all clusters */

  SCN_CHAR(scan, '}');          /* consume '}' and ';' */
  SCN_CHAR(scan, ';');          /* (end of the parameters section) */
  clset->tnew = clset->type;    /* set the cluster type */
  return 0;                     /* return 'ok' */
}  /* params() */

/*--------------------------------------------------------------------*/

static CLSET* parse (SCANNER *scan, DIMID inpcnt)
{                               /* --- parse a set of clusters */
  int     err, e;               /* error status, function result */
  DIMID   i;                    /* loop variable */
  double  *v;                   /* to resize the numeric vectors */
  CLUSTER *c;                   /* to traverse the clusters */
  CLSET   *clset;               /* created set of clusters */

  assert(scan);                 /* check the function argument */
  scn_first(scan);              /* get first token */
  clset = create(inpcnt, 0);    /* create the base structure */
  if (!clset) return NULL;      /* and initialize the variables */
  err      = scales(&clset->nst, clset->inpcnt, scan);
  if (err && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    cls_delete(clset); return NULL; }  /* read scaling parameters */
  clset->inpcnt = (clset->nst) ? nst_dim(clset->nst) : inpcnt;
  err |= e = function(clset, scan); /* read the function desc. */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    cls_delete(clset); return NULL; }
  err |= e = noise(clset, scan);    /* read the noise cluster param. */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    cls_delete(clset); return NULL; }
  err |= e = normmode(clset, scan); /* read the normalization mode */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    cls_delete(clset); return NULL; }
  err |= e = fwexp(clset, scan);    /* read the feature weight exp. */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    cls_delete(clset); return NULL; }
  err |= e = msexp(clset, scan);    /* read the membership exp. */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    cls_delete(clset); return NULL; }
  err |=     params(clset, scan);   /* read the cluster parameters */
  if (err) { cls_delete(clset); return NULL; }
  if (!clset->nst) clset->nst = nst_create(clset->inpcnt);
  clset->mat = mat_create(clset->inpcnt, clset->inpcnt);
  if (!clset->nst || !clset->mat) { cls_delete(clset); return NULL; }
  clset->cls =                  /* try to shrink the cluster vector */
  c = (CLUSTER*)realloc(clset->cls,
                           (size_t)clset->clscnt *sizeof(CLUSTER));
  v = (double*) realloc(clset->vec,
                        3 *(size_t)clset->inpcnt *sizeof(double));
  if (!v) { cls_delete(clset); return NULL; }
  clset->vec = v;               /* create the numeric vectors */
  clset->buf = v +clset->inpcnt;/* and organize them */
  for (i = 0; i < clset->clscnt; i++) {
    c[i].buf = clset->buf;      /* note the vector buffer and */
    c[i].mat = clset->mat;      /* the matrix buffer in all clusters */
  }                             /* for local computations */
  cls_type(clset, clset->type, -1);
  cls_msexp(clset,clset->msexp);/* set type and membership exponent */
  clset->resized = -1;          /* invalidate the cluster size */
  clset->setup = 0;             /* cluster set is not yet set up */
  return clset;                 /* return the created set of clusters */
}  /* parse() */

/*--------------------------------------------------------------------*/

static int header (SCANNER *scan)
{                               /* --- parse header */
  assert(scan);                 /* check the function argument */
  if ((scn_token(scan) != T_ID) /* check for a keyword */
  ||  ((strcmp(scn_value(scan), "clset")    != 0)
  &&   (strcmp(scn_value(scan), "clusters") != 0)
  &&   (strcmp(scn_value(scan), "lvqnet")   != 0)))
    SCN_ERROR(scan, E_STREXP, "clusters");
  SCN_NEXT(scan);               /* consume 'clusters' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  return 0;                     /* return 'ok' */
}  /* header() */

/*--------------------------------------------------------------------*/

static int trailer (SCANNER *scan)
{                               /* --- parse trailer */
  assert(scan);                 /* check the function argument */
  SCN_CHAR(scan, '}');          /* consume '}' */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* trailer() */

/*--------------------------------------------------------------------*/

CLSET* cls_parse (SCANNER *scan, DIMID inpcnt, int hdr)
{                               /* --- parse a cluster set */
  CLSET *clset;                 /* created cluster set */

  assert(scan);                 /* check the function arguments */
  scn_first(scan);              /* get first token */
  if (hdr && (header(scan) != 0))
    return NULL;                /* parse the header if necessary */
  clset = parse(scan, inpcnt);  /* parse the cluster set description */
  if (!clset) return NULL;      /* and then the trailer if necessary */
  if (hdr && (trailer(scan) != 0)) {
    cls_delete(clset); return NULL; }
  return clset;                 /* return the parsed cluster set */
}  /* cls_parse() */

/*--------------------------------------------------------------------*/

static int centers (MATRIX **pctrs, double **pvec,
                    DIMID inpcnt, SCANNER *scan)
{                               /* --- read cluster centers */
  DIMID  i;                     /* loop variable */
  DIMID  n;                     /* number of cluster centers */
  DIMID  m, z;                  /* size of the center vector */
  double *c;                    /* buffer for reallocation */

  assert(pctrs && pvec && scan);/* check the function arguments */
  *pctrs = NULL;                /* clear the center matrix */
  *pvec = c = NULL;             /* and the coordinate array */
  if ((scn_token(scan) != T_ID) /* check for cluster parameters */
  ||  (strcmp(scn_value(scan), "params") != 0))
    SCN_ERROR(scan, E_STREXP, "params");
  SCN_NEXT(scan);               /* consume the "params" keyword */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  n = m = z = 0;                /* initialize counters */
  while (1) {                   /* cluster center read loop */
    SCN_CHAR(scan, '{');        /* consume '{' */
    SCN_CHAR(scan, '[');        /* consume opening '[' */
    for (i = 0; 1; i++) {       /* coordinate read loop */
      if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
      if (m >= z) {             /* if coordinate vector is full */
        if (inpcnt <= 0) z += (z > BLKSIZE) ? z >> 1 : BLKSIZE;
        else z = (n +((n > BLKSIZE) ? n >> 1 : BLKSIZE)) *inpcnt;
        c = (double*)realloc(c, (size_t)z *sizeof(double));
        if (!c) SCN_ERROR(scan, E_NOMEM);
        *pvec = c;              /* enlarge the coordinate vector */
      }                         /* and set the new vector */
      c[m++] = strtod(scn_value(scan), NULL);
      SCN_NEXT(scan);           /* get and consume the coordinate */
      if (inpcnt <= 0) {        /* if input dimension is not known */
        if (scn_token(scan) != ',') { inpcnt = m; break; }
        SCN_NEXT(scan); }       /* read until no more comma */
      else {                    /* if input dimension is known */
        if (i+1 >= inpcnt) break;
        SCN_CHAR(scan, ',');    /* read required number of coords. */
      }                         /* and check for commata in between */
    }
    SCN_CHAR(scan, ']');        /* consume closing ']' */
    SCN_CHAR(scan, '}');        /* consume '}' and ';' */
    n += 1;                     /* count the cluster center */
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* consume separating comma */
  }
  SCN_CHAR(scan, '}');          /* consume '}' and ';' */
  SCN_CHAR(scan, ';');          /* (end of the parameters section) */
  *pctrs = mat_create(n,inpcnt);/* create the center matrix */
  if (!*pctrs) SCN_ERROR(scan, E_NOMEM);
  mat_init(*pctrs, MAT_FULL, c);/* initialize with read values */
  free(*pvec); *pvec = NULL;    /* and delete the read buffer */
  return 0;                     /* return 'ok' */
}  /* centers() */

/*--------------------------------------------------------------------*/

static MATRIX* cparse (SCANNER *scan, DIMID inpcnt, NSTATS **pnst)
{                               /* --- parse a set of clusters */
  int    err, e;                /* error status, function result */
  MATRIX *ctrs;                 /* created set of cluster centers */
  double *p;                    /* cluster center coordinates */

  assert(scan);                 /* check the function argument */
  scn_first(scan);              /* get first token */
  *pnst = NULL;                 /* clear the scaling parameters */
  err      = scales(pnst, inpcnt, scan);
  if (err && (scn_recover(scan, ';', 0, 0, 0) == T_EOF))
    return NULL;                /* read the scaling parameters */
  err |= e = function(NULL, scan); /* read the function description */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    if (*pnst) { nst_delete(*pnst); } return NULL; }
  err |= e = noise(NULL, scan);    /* read the noise cluster param. */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    if (*pnst) { nst_delete(*pnst); } return NULL; }
  err |= e = normmode(NULL, scan); /* read the normalization mode */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    if (*pnst) { nst_delete(*pnst); } return NULL; }
  err |= e = fwexp(NULL, scan); /* read the feature weight exponent */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    if (*pnst) { nst_delete(*pnst); } return NULL; }
  err |= e = msexp(NULL, scan); /* read the membership exponent */
  if (e   && (scn_recover(scan, ';', 0, 0, 0) == T_EOF)) {
    if (*pnst) { nst_delete(*pnst); } return NULL; }
  err |= e = centers(&ctrs, &p, inpcnt, scan);
  if (p) free(p);               /* read the cluster parameters */
  if (e) return NULL;           /* (center coordinates) */
  return ctrs;                  /* return the created cluster centers */
}  /* cparse() */

/*--------------------------------------------------------------------*/

MATRIX* cls_cparse (SCANNER *scan, DIMID inpcnt, int hdr,
                    NSTATS **pnst)
{                               /* --- parse a set of cluster centers */
  MATRIX *ctrs;                 /* created set of center clusters */

  assert(scan && pnst);         /* check the function arguments */
  scn_first(scan);              /* get first token */
  if (hdr && (header(scan) != 0))
    return NULL;                /* parse the header if necessary */
  ctrs = cparse(scan, inpcnt, pnst);
  if (!ctrs) return NULL;       /* parse the description of centers */
  if (hdr && (trailer(scan) != 0)) {
    mat_delete(ctrs); return NULL; }
  return ctrs;                  /* return the parsed cluster centers */
}  /* cls_cparse() */

#endif  /* #ifdef CLS_PARSE */
